// Alternative CFile for Non-MFC Codes
// Implemented from Microsoft Foundation Classes C++ library
// Microsoft Foundation Classes C++ library is copyrighted by Microsoft Corporation.
// To use CFile class in non-MFC codes, create CFile class and have it inherited CFileAlternative class.
// i.e. write "class CFile : public CFileAlternative {};" in your code.

#pragma once

#include <windows.h>
#include <strsafe.h>

class CFileAlternative
{
public:
	enum OpenFlags
	{
		modeRead = (int)0x00000,
		modeWrite = (int)0x00001,
		modeReadWrite = (int)0x00002,
		shareCompat = (int)0x00000,
		shareExclusive = (int)0x00010,
		shareDenyWrite = (int)0x00020,
		shareDenyRead = (int)0x00030,
		shareDenyNone = (int)0x00040,
		modeNoInherit = (int)0x00080,
		modeCreate = (int)0x01000,
		modeNoTruncate = (int)0x02000,
		typeText = (int)0x04000,
		typeBinary = (int)0x08000,
		osNoBuffer = (int)0x10000,
		osWriteThrough = (int)0x20000,
		osRandomAccess = (int)0x40000,
		osSequentialScan = (int)0x80000,
	};
	enum Attribute
	{
		normal = 0x00,
		readOnly = 0x01,
		hidden = 0x02,
		system = 0x04,
		volume = 0x08,
		directory = 0x10,
		archive = 0x20
	};
	enum SeekPosition
	{
		begin = 0x0,
		current = 0x1,
		end = 0x2
	};
	enum BufferCommand
	{
		bufferRead,
		bufferWrite,
		bufferCommit,
		bufferCheck
	};
	enum BufferFlags 
	{ 
		bufferDirect = 0x01,
		bufferBlocking = 0x02
	};
	HANDLE m_hFile;
	CFileAlternative()
	{
		m_hFile = INVALID_HANDLE_VALUE;
		m_bCloseOnDelete = FALSE;
	}
	CFileAlternative(HANDLE hFile)
	{
		m_hFile = hFile;
		m_bCloseOnDelete = FALSE;
	}
	CFileAlternative(LPCTSTR lpszFileName, UINT nOpenFlags)
	{
		m_hFile = INVALID_HANDLE_VALUE;
		Open(lpszFileName, nOpenFlags, NULL);
	}
	virtual ~CFileAlternative()
	{
		if(m_hFile != INVALID_HANDLE_VALUE && m_bCloseOnDelete)
			Close();
	}
	virtual CFileAlternative* Duplicate() const
	{
		CFileAlternative* pFile;
		HANDLE hFile;
		pFile = new CFileAlternative();
		if(!::DuplicateHandle(::GetCurrentProcess(), m_hFile, ::GetCurrentProcess(), &hFile, 0, FALSE, DUPLICATE_SAME_ACCESS))
		{
			delete pFile;
			return NULL;
		}
		pFile->m_hFile = hFile;
		pFile->m_bCloseOnDelete = m_bCloseOnDelete;
		return pFile;
	}
	virtual BOOL Open(LPCTSTR lpszFileName, UINT nOpenFlags, void* pException)
	{
		DWORD dwAccess;
		DWORD dwShareMode;
		SECURITY_ATTRIBUTES sa;
		DWORD dwCreateFlag;
		DWORD dwFlags;
		HANDLE hFile;
		nOpenFlags &= ~(UINT)typeBinary;
		m_bCloseOnDelete = FALSE;
		m_hFile = INVALID_HANDLE_VALUE;
		if(lpszFileName != NULL && SUCCEEDED(StringCchLength(lpszFileName, _MAX_PATH, NULL)))
		{
		}
		else
		{
			if(pException != NULL)
			{
			}
			return FALSE;
		}
		dwAccess = 0;
		switch(nOpenFlags & 3)
		{
		case modeRead:
			dwAccess = GENERIC_READ;
			break;
		case modeWrite:
			dwAccess = GENERIC_WRITE;
			break;
		case modeReadWrite:
			dwAccess = GENERIC_READ | GENERIC_WRITE;
			break;
		default:
			break;
		}
		dwShareMode = 0;
		switch(nOpenFlags & 0x70)
		{
		default:
		case shareCompat:
		case shareExclusive:
			dwShareMode = 0;
			break;
		case shareDenyWrite:
			dwShareMode = FILE_SHARE_READ;
			break;
		case shareDenyRead:
			dwShareMode = FILE_SHARE_WRITE;
			break;
		case shareDenyNone:
			dwShareMode = FILE_SHARE_WRITE | FILE_SHARE_READ;
			break;
		}
		sa.nLength = sizeof(sa);
		sa.lpSecurityDescriptor = NULL;
		sa.bInheritHandle = (nOpenFlags & modeNoInherit) == 0;
		if(nOpenFlags & modeCreate)
		{
			if(nOpenFlags & modeNoTruncate)
				dwCreateFlag = OPEN_ALWAYS;
			else
				dwCreateFlag = CREATE_ALWAYS;
		}
		else
			dwCreateFlag = OPEN_EXISTING;
		dwFlags = FILE_ATTRIBUTE_NORMAL;
		if(nOpenFlags & osNoBuffer)
			dwFlags |= FILE_FLAG_NO_BUFFERING;
		if(nOpenFlags & osWriteThrough)
			dwFlags |= FILE_FLAG_WRITE_THROUGH;
		if(nOpenFlags & osRandomAccess)
			dwFlags |= FILE_FLAG_RANDOM_ACCESS;
		if(nOpenFlags & osSequentialScan)
			dwFlags |= FILE_FLAG_SEQUENTIAL_SCAN;
		hFile = ::CreateFile(lpszFileName, dwAccess, dwShareMode, &sa, dwCreateFlag, dwFlags, NULL);
		if(hFile == INVALID_HANDLE_VALUE)
			return FALSE;
		m_hFile = hFile;
		m_bCloseOnDelete = TRUE;
		return TRUE;
	}
	virtual UINT Read(void* lpBuf, UINT nCount)
	{
		DWORD dwRead;
		if(nCount == 0)
			return 0;
		if(!::ReadFile(m_hFile, lpBuf, nCount, &dwRead, NULL))
			return 0;
		return (UINT)dwRead;
	}
	virtual void Write(const void* lpBuf, UINT nCount)
	{
		DWORD nWritten;
		if(nCount == 0)
			return;
		if(!::WriteFile(m_hFile, lpBuf, nCount, &nWritten, NULL))
			return;
		if(nWritten != nCount)
			return;
	}
	virtual ULONGLONG Seek(LONGLONG lOff, UINT nFrom)
	{
		LARGE_INTEGER liOff;
		liOff.QuadPart = lOff;
		liOff.LowPart = ::SetFilePointer(m_hFile, liOff.LowPart, &liOff.HighPart, (DWORD)nFrom);
		if(liOff.LowPart == (DWORD)-1)
		{
			if(::GetLastError() != NO_ERROR)
				return 0;
		}
		return liOff.QuadPart;
	}
	virtual ULONGLONG GetPosition() const
	{
		LARGE_INTEGER liPos;
		liPos.QuadPart = 0;
		liPos.LowPart = ::SetFilePointer(m_hFile, liPos.LowPart, &liPos.HighPart, FILE_CURRENT);
		if(liPos.LowPart == (DWORD)-1)
		{
			if(::GetLastError() != NO_ERROR)
				return 0;
		}
		return liPos.QuadPart;
	}
	virtual void Flush()
	{
		if(m_hFile == INVALID_HANDLE_VALUE)
			return;
		if(!::FlushFileBuffers(m_hFile))
			return;
	}
	virtual void Close()
	{
		BOOL bError;
		bError = FALSE;
		if(m_hFile != INVALID_HANDLE_VALUE)
			bError = !::CloseHandle(m_hFile);
		m_hFile = INVALID_HANDLE_VALUE;
		m_bCloseOnDelete = FALSE;
		if(bError)
			return;
	}
	virtual void Abort()
	{
		if(m_hFile != INVALID_HANDLE_VALUE)
		{
			::CloseHandle(m_hFile);
			m_hFile = INVALID_HANDLE_VALUE;
		}
	}
#if defined(_X86_) || defined(_AMD64_)
	virtual void LockRange(ULONGLONG dwPos, ULONGLONG dwCount)
	{
		ULARGE_INTEGER liPos;
		ULARGE_INTEGER liCount;
		liPos.QuadPart = dwPos;
		liCount.QuadPart = dwCount;
		if(!::LockFile(m_hFile, liPos.LowPart, liPos.HighPart, liCount.LowPart, liCount.HighPart))
			return;
	}
	virtual void UnlockRange(ULONGLONG dwPos, ULONGLONG dwCount)
	{
		ULARGE_INTEGER liPos;
		ULARGE_INTEGER liCount;
		liPos.QuadPart = dwPos;
		liCount.QuadPart = dwCount;
		if(!::UnlockFile(m_hFile, liPos.LowPart, liPos.HighPart, liCount.LowPart, liCount.HighPart))
			return;
	}
#endif
	virtual void SetLength(ULONGLONG dwNewLen)
	{
		Seek(dwNewLen, (UINT)begin);
		if(!::SetEndOfFile(m_hFile))
			return;
	}
	virtual ULONGLONG GetLength() const
	{
		ULARGE_INTEGER liSize;
		liSize.LowPart = ::GetFileSize(m_hFile, &liSize.HighPart);
		if(liSize.LowPart == INVALID_FILE_SIZE)
		{
			if(::GetLastError() != NO_ERROR)
				return 0;
		}
		return liSize.QuadPart;
	}
	virtual UINT GetBufferPtr(UINT nCommand, UINT nCount, void** ppBufStart, void** ppBufMax)
	{
		return 0;
	}
protected:
	BOOL m_bCloseOnDelete;
};

