#pragma once

#define KANJI_MAX_KANA					16
#define KANJI_MAX_KANJI					16
#define KANJI_MAX_RELATION				40
#define KANJI_MAX_WORD					64
#define KANJI_MAX_BUNSETSU				64
#define KANJI_INVALID_CHAR				0xffff
#define KANJI_MAX_HIGHEST_BUNSETSU		2
#define KANJI_MAX_HIGHEST_ENUM			24
#define KANJI_MAX_PROOF					16

struct KANJI_PREFERENCE
{
	DWORD Timeout;
	DWORD TimeoutCount;
	LONG PointThreshold1;
	LONG PointThreshold2;
	LONG PointThreshold3;
	LONG PointHistory;
	LONG PointChangeBunsetsu;
	LONG PointFuzokugo;
	LONG PointMultipleWord;
	LONG PointRelation;
	LONG PointNotBegin;
	LONG PointNotEnd;
	LONG PointEndYougen;
	LONG PointEndFuzokugo;
	LONG PointPreferBegin;
	LONG PointPreferEnd;
	LONG PointSequenceKatsuyou;
	LONG PointSequenceMeishi;
	LONG PointSequenceJinmei;
	LONG PointSequenceSuushi;
	LONG PointMeishiJoshi;
	LONG PointKatsuyouMeishi;
	LONG PointSettoujiJinmei;
	LONG PointJinmeiSetsubiji;
	LONG PointMultipleJiritsugo;
	LONG PointDifferentSuushi;
	LONG PointRelationKatsuyou;
	LONG PointMismatch;
	LONG PointMissPrev;
	LONG PointMissNext;
	LONG PointWrongKatsuyou;
	LONG PointMultipleChar;
	LONG PointMultipleBunsetsu;
	LONG PointDefaultLearning;
	DWORD MinComplement;
	DWORD MaxProofCount;
};

struct KANJI_DICTIONARY_HEADER
{
	DWORD Magic;
	DWORD Version;
	DWORD Count;
	DWORD Reserved;
	TCHAR Description[32];
	TCHAR Copyright[32];
};

struct KANJI_DICTIONARY
{
	DWORD Type;
	DWORD PrevType;
	DWORD NextType;
	LONG Point;
	TCHAR Kana[KANJI_MAX_KANA];
	TCHAR Kanji[KANJI_MAX_KANJI];
	TCHAR Relation[KANJI_MAX_RELATION];
};

struct KANJI_BUNSETSU
{
	KANJI_DICTIONARY* pKanji[KANJI_MAX_WORD];
	TCHAR Kana[KANJI_MAX_WORD];
};

#define KANJI_COMMON_MASK				0x000fffff
#define KANJI_TYPE_MASK					0x00ffffff
#define KANJI_TYPE_NOT_BEGIN			0x00100000
#define KANJI_TYPE_NOT_END				0x00200000
#define KANJI_TYPE_NO_RELATION			0x00400000
#define KANJI_TYPE_NO_COUNT				0x00800000
#define KANJI_TYPE_DOUSHI				0x00000001
#define KANJI_TYPE_KEIYOUSHI			0x00000002
#define KANJI_TYPE_KEIYOUMEISHI			0x00000014
#define KANJI_TYPE_KATSUYOU				(0x00000008 | KANJI_TYPE_NOT_BEGIN | KANJI_TYPE_NO_RELATION)
#define KANJI_TYPE_MEISHI				0x00000010
#define KANJI_TYPE_JOSHI				(0x00000020 | KANJI_TYPE_NOT_BEGIN | KANJI_TYPE_NO_RELATION)
#define KANJI_TYPE_FUKUSHI				(0x00000040 | KANJI_TYPE_NO_RELATION)
#define KANJI_TYPE_KOYUUMEISHI			0x00000090
#define KANJI_TYPE_JINMEI				0x00000190
#define KANJI_TYPE_GAIRAIGO				0x00000210
#define KANJI_TYPE_KUTOUTEN				0x00000400
#define KANJI_TYPE_SUUSHI				0x00000800
#define KANJI_TYPE_SAHENDOUSHI			0x00001000
#define KANJI_TYPE_ZAHENDOUSHI			0x00002000
#define KANJI_TYPE_SETTOUJI				(0x00000010 | KANJI_TYPE_NO_RELATION | KANJI_TYPE_NO_COUNT)
#define KANJI_TYPE_SETSUBIJI			(0x00004000 | KANJI_TYPE_NO_RELATION)
#define KANJI_TYPE_BUNSETSU				0x00008000
#define KANJI_TYPE_SUUSHI_MASK			0x000f0000
#define KANJI_TYPE_SUUSHI_DEFAULT		(0x00000000 | KANJI_TYPE_SUUSHI)
#define KANJI_TYPE_SUUSHI_ARABIC		(0x00010000 | KANJI_TYPE_SUUSHI)
#define KANJI_TYPE_SUUSHI_KANJI			(0x00020000 | KANJI_TYPE_SUUSHI)
#define KANJI_TYPE_SUUSHI_DAIJI			(0x00040000 | KANJI_TYPE_SUUSHI)
#define KANJI_TYPE_SUUSHI_KANJI_DAIJI	(0x00060000 | KANJI_TYPE_SUUSHI)
#define KANJI_TYPE_SUUSHI_ALL			(0x000f0000 | KANJI_TYPE_SUUSHI)
#define KANJI_TYPE_YOUGEN				0x00000007
#define KANJI_TYPE_JIRITSUGO			0x00008857
#define KANJI_TYPE_FUZOKUGO				0x00007028
#define KANJI_TYPE_DEFAULT				KANJI_TYPE_MEISHI
#define KANJI_TYPE_ALL					0x0000f87f
#define KANJI_PREV_MASK					0x0f0fffff
#define KANJI_PREV_NEEDED				0x01000000
#define KANJI_PREV_PREFER_BEGIN			0x02000000
#define KANJI_PREV_DOUSHI				(KANJI_TYPE_DOUSHI & KANJI_PREV_MASK)
#define KANJI_PREV_KEIYOUSHI			((KANJI_TYPE_KEIYOUSHI | (KANJI_TYPE_KEIYOUMEISHI ^ KANJI_TYPE_MEISHI) | KANJI_TYPE_KATSUYOU) & KANJI_PREV_MASK)
#define KANJI_PREV_KATSUYOU				(KANJI_TYPE_KATSUYOU & KANJI_PREV_MASK)
#define KANJI_PREV_YOUGEN				((KANJI_TYPE_YOUGEN | KANJI_TYPE_KATSUYOU | KANJI_TYPE_SAHENDOUSHI | KANJI_PREV_NEEDED) & KANJI_PREV_MASK)
#define KANJI_PREV_SETSUBIJI			((KANJI_TYPE_MEISHI | KANJI_PREV_NEEDED) & KANJI_PREV_MASK)
#define KANJI_PREV_SETSUBIJI_SUUSHI		((KANJI_TYPE_SUUSHI | KANJI_PREV_NEEDED) & KANJI_PREV_MASK)
#define KANJI_PREV_RENTAI				((KANJI_TYPE_MEISHI | KANJI_TYPE_JOSHI | KANJI_TYPE_FUKUSHI) & KANJI_PREV_MASK)
#define KANJI_PREV_RENDAKU				((KANJI_TYPE_DOUSHI | KANJI_TYPE_KATSUYOU | KANJI_TYPE_MEISHI | KANJI_TYPE_SUUSHI | KANJI_TYPE_SAHENDOUSHI | KANJI_PREV_NEEDED) & KANJI_PREV_MASK)
#define KANJI_PREV_KEISHOU				(((KANJI_TYPE_JINMEI ^ KANJI_TYPE_KOYUUMEISHI) | KANJI_PREV_NEEDED) & KANJI_PREV_MASK)
#define KANJI_PREV_TANI					((KANJI_TYPE_MEISHI | KANJI_TYPE_SUUSHI) & KANJI_PREV_MASK)
#define KANJI_PREV_SAHENGOKAN			((KANJI_TYPE_MEISHI | KANJI_PREV_NEEDED) & KANJI_PREV_MASK)
#define KANJI_PREV_DEFAULT				(KANJI_TYPE_ALL & KANJI_PREV_MASK)
#define KANJI_PREV_ALL					(KANJI_TYPE_ALL & KANJI_PREV_MASK)
#define KANJI_NEXT_MASK					0xf00fffff
#define KANJI_NEXT_NEEDED				0x10000000
#define KANJI_NEXT_PREFER_END			0x20000000
#define KANJI_NEXT_KATSUYOU				((KANJI_TYPE_KATSUYOU | KANJI_NEXT_NEEDED) & KANJI_NEXT_MASK)
#define KANJI_NEXT_MEISHI				((KANJI_TYPE_YOUGEN | KANJI_TYPE_MEISHI | KANJI_TYPE_JOSHI | KANJI_TYPE_FUKUSHI | KANJI_TYPE_SUUSHI | KANJI_TYPE_SETSUBIJI | KANJI_TYPE_BUNSETSU) & KANJI_NEXT_MASK)
#define KANJI_NEXT_KEIYOUMEISHI			((KANJI_TYPE_YOUGEN | KANJI_TYPE_KATSUYOU | KANJI_TYPE_MEISHI | KANJI_TYPE_JOSHI | KANJI_TYPE_FUKUSHI | KANJI_TYPE_SUUSHI | KANJI_TYPE_SETSUBIJI | KANJI_TYPE_BUNSETSU) & KANJI_NEXT_MASK)
#define KANJI_NEXT_SETTOUJI				((KANJI_TYPE_MEISHI | KANJI_NEXT_NEEDED) & KANJI_NEXT_MASK)
#define KANJI_NEXT_SETTOUJI_DOUSHI		((KANJI_TYPE_DOUSHI | KANJI_NEXT_NEEDED) & KANJI_NEXT_MASK)
#define KANJI_NEXT_SETTOUJI_YOUGEN		((KANJI_TYPE_YOUGEN | KANJI_NEXT_NEEDED) & KANJI_NEXT_MASK)
#define KANJI_NEXT_SETTOUJI_SUUSHI		((KANJI_TYPE_SUUSHI | KANJI_NEXT_NEEDED) & KANJI_NEXT_MASK)
#define KANJI_NEXT_SUUSHI				((KANJI_TYPE_MEISHI | KANJI_TYPE_JOSHI | KANJI_TYPE_SUUSHI | KANJI_TYPE_SETSUBIJI) & KANJI_NEXT_MASK)
#define KANJI_NEXT_SUUSHI_ONBIN			((KANJI_TYPE_MEISHI | KANJI_TYPE_JOSHI | KANJI_TYPE_SUUSHI | KANJI_TYPE_SETSUBIJI | KANJI_NEXT_NEEDED) & KANJI_NEXT_MASK)
#define KANJI_NEXT_RENYOU				((KANJI_TYPE_YOUGEN | KANJI_TYPE_KATSUYOU | KANJI_TYPE_MEISHI | KANJI_TYPE_JOSHI | KANJI_TYPE_FUKUSHI | KANJI_TYPE_SUUSHI | KANJI_TYPE_BUNSETSU) & KANJI_NEXT_MASK)
#define KANJI_NEXT_RENTAI				((KANJI_TYPE_YOUGEN | KANJI_TYPE_MEISHI | KANJI_TYPE_JOSHI | KANJI_TYPE_FUKUSHI | KANJI_TYPE_SUUSHI | KANJI_TYPE_BUNSETSU | KANJI_NEXT_PREFER_END) & KANJI_NEXT_MASK)
#define KANJI_NEXT_SHUUSHI				((KANJI_TYPE_JOSHI | KANJI_TYPE_KUTOUTEN | KANJI_NEXT_PREFER_END) & KANJI_NEXT_MASK)
#define KANJI_NEXT_ONBIN				((KANJI_TYPE_YOUGEN | KANJI_TYPE_KATSUYOU | KANJI_TYPE_MEISHI | KANJI_TYPE_SUUSHI | KANJI_NEXT_NEEDED) & KANJI_NEXT_MASK)
#define KANJI_NEXT_SHUUJOSHI			((KANJI_TYPE_KUTOUTEN | KANJI_NEXT_PREFER_END) & KANJI_NEXT_MASK)
#define KANJI_NEXT_SAHENMEISHI			((KANJI_TYPE_YOUGEN | KANJI_TYPE_MEISHI | KANJI_TYPE_JOSHI | KANJI_TYPE_FUKUSHI | KANJI_TYPE_SUUSHI | KANJI_TYPE_SETSUBIJI | KANJI_TYPE_BUNSETSU | KANJI_TYPE_SAHENDOUSHI) & KANJI_NEXT_MASK)
#define KANJI_NEXT_ZAHENMEISHI			((KANJI_TYPE_YOUGEN | KANJI_TYPE_MEISHI | KANJI_TYPE_JOSHI | KANJI_TYPE_FUKUSHI | KANJI_TYPE_SUUSHI | KANJI_TYPE_SETSUBIJI | KANJI_TYPE_BUNSETSU | KANJI_TYPE_ZAHENDOUSHI) & KANJI_NEXT_MASK)
#define KANJI_NEXT_SAHENDOUSHI			((KANJI_TYPE_SAHENDOUSHI | KANJI_NEXT_NEEDED) & KANJI_NEXT_MASK)
#define KANJI_NEXT_ZAHENDOUSHI			((KANJI_TYPE_ZAHENDOUSHI | KANJI_NEXT_NEEDED) & KANJI_NEXT_MASK)
#define KANJI_NEXT_BUNSETSU				(KANJI_TYPE_JIRITSUGO & KANJI_NEXT_MASK)
#define KANJI_NEXT_DEFAULT				(KANJI_TYPE_ALL & KANJI_NEXT_MASK)
#define KANJI_NEXT_ALL					(KANJI_TYPE_ALL & KANJI_NEXT_MASK)

#define KANJI_POINT_NOT_GENERAL			-512
#define KANJI_POINT_DEFAULT				0
#define KANJI_POINT_GENERAL				128
#define KANJI_POINT_FREQUENT			256
#define KANJI_POINT_PRECEDED			512
#define KANJI_POINT_NOT_FOUND			-65536

struct KANJI_CANDIDATE
{
	KANJI_DICTIONARY* pData;
	LONG Point;
};

struct KANJI_BUNSETSU_CANDIDATE
{
	KANJI_BUNSETSU Data;
	LONG Point;
};

#define KANJI_KEY_ENTER					1
#define KANJI_KEY_CONVERT				2
#define KANJI_KEY_BACK					3
#define KANJI_KEY_DELETE				4
#define KANJI_KEY_LEFT					5
#define KANJI_KEY_RIGHT					6
#define KANJI_KEY_PREV					7
#define KANJI_KEY_NEXT					8
#define KANJI_KEY_COMBINE				9

#define KANJI_INPUT_NONE				0
#define KANJI_INPUT_TYPING				1
#define KANJI_INPUT_CANDIDATE			2
#define KANJI_INPUT_COMBINATION			3
#define KANJI_INPUT_COMPLEMENT			4

extern HWND g_hWnd;
extern KANJI_PREFERENCE g_Preference;
extern KANJI_DICTIONARY* g_pDictionary;
extern DWORD g_dwDictionary;
extern DWORD g_dwMaxDictionary;
extern KANJI_DICTIONARY* g_pInputDictionary;
extern DWORD g_dwInputDictionary;
extern DWORD g_dwMaxInputDictionary;
extern KANJI_DICTIONARY g_TempDictionary[KANJI_MAX_WORD];
extern DWORD g_dwTempDictionary;
extern BOOL g_bTimeout;
extern DWORD g_StartTime;

void ResetTimeout();
BOOL CheckTimeout();
LONG CompareBunsetsu(KANJI_BUNSETSU* p1, KANJI_BUNSETSU* p2);
DWORD FindKanjiHead(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* Kana, DWORD Count);
KANJI_DICTIONARY* FindKanji(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* Kana, DWORD Count, DWORD* pLastFound);
LONG GetRelationPoint(KANJI_DICTIONARY* p1, KANJI_DICTIONARY* p2);
LONG GetSentenceRelationPoint(KANJI_DICTIONARY** pp, DWORD Count, BOOL bCheckBegin, BOOL bCheckEnd);
LONG GetSentenceRelationPoint(KANJI_DICTIONARY** pp, DWORD Count);
LONG CheckSentencePoint(KANJI_DICTIONARY** pp, DWORD Count);
LONG GetBunsetsuPoint(KANJI_DICTIONARY** pp, DWORD Count);
LONG CheckBunsetsuPoint(KANJI_DICTIONARY** pp, DWORD Count);
DWORD GetLongestBunsetsu(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* pString, DWORD Count, KANJI_BUNSETSU* pResult, BOOL bJoshi, LONG Threshold);
DWORD EnumBunsetsu(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* pString, DWORD Count, KANJI_BUNSETSU* pResult, DWORD dwResult, BOOL bJoshi, LONG Threshold);
BOOL GetFirstBunsetsu(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* pString, DWORD Count, KANJI_BUNSETSU* pResult, BOOL bJoshi, LONG Threshold);
DWORD GetKanjiFromBunsetsu(KANJI_BUNSETSU* pResult, KANJI_DICTIONARY** ppResult);
LONG GetBunsetsuRelationPoint(KANJI_BUNSETSU* p, DWORD Count, BOOL bCheckBegin, BOOL bCheckEnd);
LONG CheckBunsetsuRelationPoint(KANJI_BUNSETSU* p, DWORD Count);
LONG CheckProofKanaToBunsetsu(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* pString, KANJI_BUNSETSU* pResult, LONG Threshold);
LONG ConvertKanaToBunsetsu(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* pString, DWORD Count, KANJI_BUNSETSU* pResult, LONG Threshold);
LONG ConvertKanaToBunsetsu2Longest(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* pString, DWORD Count, KANJI_BUNSETSU* pResult, LONG Threshold);
LONG ConvertKanaToBunsetsuHighest(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* pString, DWORD Count, KANJI_BUNSETSU* pResult, LONG Threshold);
void ConvertKanaToBunsetsu(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* pString, DWORD Count, KANJI_BUNSETSU* pResult);
DWORD EnumTempBunsetsu(TCHAR* Kana, KANJI_BUNSETSU* pResult, DWORD dwResult);
LONG CheckProofKanaToBunsetsu(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* pString, KANJI_BUNSETSU* pResult, LONG Threshold);
void ProofKanaToBunsetsu(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* pString, KANJI_BUNSETSU* pResult);
void GetKana(KANJI_DICTIONARY** ppResult, TCHAR* pString);
void GetKanaFromBunsetsu(KANJI_BUNSETSU* pResult, TCHAR* pString);
void ConvertInputToKana(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* pString, DWORD Count, TCHAR* pOutput);
void MergeDictionaryData(KANJI_DICTIONARY* pBuf, KANJI_DICTIONARY* pData, DWORD Count);
void SortDictionary(KANJI_DICTIONARY* pData, DWORD Count);
void MergeCandidateData(KANJI_CANDIDATE* pBuf, KANJI_CANDIDATE* pData, DWORD Count);
void SortCandidate(KANJI_CANDIDATE* pData, DWORD Count);
void MergeBunsetsuCandidateData(KANJI_BUNSETSU_CANDIDATE* pBuf, KANJI_BUNSETSU_CANDIDATE* pData, DWORD Count);
void SortBunsetsuCandidate(KANJI_BUNSETSU_CANDIDATE* pData, DWORD Count);
void MergeDictionaryDataPointerPoint(KANJI_DICTIONARY** ppBuf, KANJI_DICTIONARY** ppData, DWORD Count);
DWORD EnumKanjiCandidate(KANJI_DICTIONARY* pData, DWORD dwData, KANJI_DICTIONARY** ppKanji, DWORD Offset, KANJI_CANDIDATE* pOutput);
DWORD EnumBunsetsuCandidate(KANJI_DICTIONARY* pData, DWORD dwData, KANJI_BUNSETSU* pBunsetsu, DWORD Offset, KANJI_BUNSETSU_CANDIDATE* pOutput);
DWORD EnumBunsetsuComplement(KANJI_DICTIONARY* pData, DWORD dwData, TCHAR* pString, KANJI_BUNSETSU_CANDIDATE* pOutput);
BOOL SetDictionarySize(DWORD Count);
void SetDictionaryDataTrim(KANJI_DICTIONARY* pData, DWORD Count);
void SetDictionaryDataToTop(KANJI_DICTIONARY* pData, DWORD Count);
void SetDictionaryDataToKanaTop(KANJI_DICTIONARY** ppData, DWORD Count);
void TrimDictionaryData();
void SetDictionaryData(DWORD Type, DWORD PrevType, DWORD NextType, LONG Point, LPCTSTR Kana, LPCTSTR Kanji, LPCTSTR Relation);
void SetDictionaryData(DWORD Type, DWORD PrevType, DWORD NextType, LPCTSTR Kana, LPCTSTR Kanji, LPCTSTR Relation);
void SetDictionaryData(DWORD Type, DWORD NextType, LPCTSTR Kana, LPCTSTR Kanji, LPCTSTR Relation);
void SetDictionaryData(DWORD Type, LPCTSTR Kana, LPCTSTR Kanji, LPCTSTR Relation);
void SetDictionaryDummyData(DWORD Type, DWORD NextType, LPCTSTR Kana, LPCTSTR Kanji);
void SetDictionaryDummyData(DWORD Type, DWORD NextType, LPCTSTR Kana);
void SetDictionaryDummyData(DWORD Type, LPCTSTR Kana);
BOOL SetInputDictionarySize(DWORD Count);
void SetInputDictionaryData(LPCTSTR Input, LPCTSTR Kana);
void GetKanjiTypeString(KANJI_DICTIONARY* pKanji, TCHAR* pOut);
void GetBunsetsuTypeString(KANJI_BUNSETSU* pBunsetsu, TCHAR* pOut);

